/*
 * Copyright 2020 Arne Vajh�j.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

#include <iostream>
#include <iomanip>
#include <vector>
#include <map>
#include <algorithm>
#include <string>
#include <cstdlib>

using namespace std;

#include "clockstack.h"

struct element
{
    string func;
    TIMECOUNT_T totaltime;
    TIMECOUNT_T usedtime;
};

static vector<string> callstack;
static vector<TIMECOUNT_T> callcounters;
static map<string,element> stats;
static string lastfunc;

void clockstack_enter(const char *func, TIMECOUNT_T timecount)
{
    // push function name and time counter on stacks
    callstack.push_back(string(func));
    callcounters.push_back(timecount);
    // create statistics element if it does not exist
    if(stats[func].func != func)
    {
        element elm;
        elm.func = func;
        elm.totaltime = 0;
        elm.usedtime = 0;
        stats[func] = elm;
    }
}

void clockstack_exit(const char *func, TIMECOUNT_T timecount)
{
    // check if consistent call stack
    if(callstack.empty())
    {
        cerr << "More exits than enters" << endl;
        exit(EXIT_FAILURE);
    }
    string expfunc = callstack.back();
    if(func != expfunc)
    {
        cerr << "Actual exit function name does not match expected exit function name" << endl;
        exit(EXIT_FAILURE);
    }
    // time spent since enter
    TIMECOUNT_T dtime = timecount - callcounters.back();
    // pop stack
    callstack.pop_back();
    callcounters.pop_back();
    // update statistics for current function
    stats[func].totaltime += dtime;
    stats[func].usedtime += dtime;
    // update statistics for caller function (deduct time spent here from its used time)
    if(callstack.rbegin() != callstack.rend())
    {
        stats[*callstack.rbegin()].usedtime -= dtime;
    }
    // save outermost function (outermost function total time becomes 100%)
    if(callstack.empty()) 
    {
        lastfunc = func;
    }
}

static bool bytotal(element e1, element e2)
{
    return e1.totaltime > e2.totaltime;
}

static bool byused(element e1, element e2)
{
    return e1.usedtime > e2.usedtime;
}

void clockstack_dump(void)
{
    TIMECOUNT_T total = stats[lastfunc].totaltime;
    vector<element> temp;
    for(map<string,element>::iterator it = stats.begin(); it != stats.end(); it++)
    {
        temp.push_back(it->second);
    }
    cout.setf(ios::fixed);
    cout << "Total time:" << endl;
    sort(temp.begin(), temp.end(), bytotal);
    for(vector<element>::iterator it = temp.begin(); it != temp.end(); it++)
    {
        cout << it->func << ": " << setprecision(1) << (it->totaltime * 100.0 / total) << "%" << endl;
    }
    cout << "Used time:" << endl;
    sort(temp.begin(), temp.end(), byused);
    for(vector<element>::iterator it = temp.begin(); it != temp.end(); it++)
    {
        cout << it->func << ": " << setprecision(1) << (it->usedtime * 100.0 / total) << "%" << endl;
    }
}
